-- Topal: GPG/GnuPG and Alpine/Pine integration
-- Copyright (C) 2001--2008  Phillip J. Brooke
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License version 3 as
-- published by the Free Software Foundation.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.

with Ada.IO_Exceptions;
with Ada.Strings.Fixed;
with Ada.Strings.Maps;
with Ada.Strings.Maps.Constants;
with Ada.Strings.Unbounded;
with Ada.Text_IO;
with Expanding_Array;
with Externals;             use Externals;
with Externals.GPG;
with Menus;                 use Menus;
with Misc;                  use Misc;
with Readline;

package body Keys is

   Key_List_Initial_Size : constant Positive := 200;
   Key_Head_Length       : constant Positive := 72;

   subtype Key_Head is String(1..Key_Head_Length);
   package Key_Head_Array_Package
   is new Expanding_Array (Item => Key_Head);
   package KHA renames Key_Head_Array_Package;

   subtype Key_Head_Array is Key_Head_Array_Package.Big_Array;

   type Menu_Array_Ptr is access all Keylist_Menus.MNA;

   function Match_PR (KP : Key_Properties;
                      KR : Key_Roles) return Boolean is
   begin
      case KR is
         when Any       => return True;
         when Encrypter => return KP(Encrypter);
         when Signer    => return KP(Signer);
         when Both      => return KP(Encrypter) and KP(Signer);
      end case;
   end Match_PR;

   -- Is the key usable?
   function Usable (KP : Key_Properties) return Boolean is
   begin
      return not (KP(Invalid) or KP(Disabled) or KP(Revoked) or KP(Expired));
   end Usable;

   -- Given a line from Listkey, return a processed version.
   procedure Process_Key (L  : in     String;
                          P  :    out UBS;
                          KP :    out Key_Properties) is
      LS : UBS_Array := Split_GPG_Colons(L);
      use Ada.Strings.Fixed;
   begin
      P := NullUBS;
      KP := Key_Properties'(others => False);
      if LS'Length >= 1
        and then ToStr(LS(LS'First)) = "pub" then
         if LS'Length >= 2 then
            declare
               S : constant String := ToStr(LS(LS'First+1));
            begin
               if Count(S, "i") = 1 then
                  KP(Invalid) := True;
               end if;
               if Count(S, "d") = 1 then
                  KP(Disabled) := True;
               end if;
               if Count(S, "r") = 1 then
                  KP(Revoked) := True;
               end if;
               if Count(S, "e") = 1 then
                  KP(Expired) := True;
               end if;
            end;
         end if;
         if LS'Length >= 12 then
            declare
               S : constant String := ToStr(LS(LS'First+11));
            begin
               if Count(S, "D") = 1 then
                  KP(Disabled) := True;
               end if;
               if Count(S, "E") = 1 then
                  KP(Encrypter) := True;
               end if;
               if Count(S, "S") = 1 then
                  KP(Signer) := True;
               end if;
            end;
         end if;
         declare
            use type UBS;
         begin
            if KP(Invalid) then
               P := P & ToUBS("[Invalid]");
            end if;
            if KP(Disabled) then
               P := P & ToUBS("[Disabled]");
            end if;
            if KP(Revoked) then
               P := P & ToUBS("[Revoked]");
            end if;
            if KP(Expired) then
               P := P & ToUBS("[Expired]");
            end if;
            if LS'Length >= 5 then
               -- Append the last eight characters of the key ID.
               declare
                  KS : constant String := ToStr(LS(LS'First+4));
                  F  : Integer;
               begin
                  F := KS'Last - 7;
                  if F < KS'First then
                     F := KS'First;
                  end if;
                  P := P & KS(F..KS'Last);
               end;
            end if;
            if LS'Length >= 6 then
               -- Append the start date.
               P := P & ToUBS(" ") & LS(LS'First+5);
            end if;
            if LS'Length >= 10 then
               -- Append the user name.
               P := P & ToUBS(" ") & LS(LS'First+9);
            end if;
         end;
      else
         P := ToUBS("??:" & L);
      end if;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Process_Key");
         raise;
   end Process_Key;

   -- Add a key to the list....
   procedure Add_Key (Key  : in     UBS;
                      List : in out Key_List;
                      Role : in     Key_Roles) is
      Match        : Boolean := False;
      Key_Filename : constant String := Temp_File_Name("addkey");
      P            : UBS;
      KP           : Key_Properties;
      File_Handle  : Ada.Text_IO.File_Type;
      First_Line   : UBS;
   begin
      Debug("+Add_Key: `" & ToStr(Key) & "'");
      -- Puke if this is not a fingerprint....
      declare
         KS : constant String := ToStr(Key);
         use Ada.Strings.Maps;
      begin
         for I in KS'Range loop
            if not Is_In(KS(I),
                         Ada.Strings.Maps.Constants.Hexadecimal_Digit_Set) then
               Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                                    "Expected fingerprint, but working on `"&
                                    KS &"'");
               raise Fingerprint_Expected;
            end if;
         end loop;
      end;
      -- Is this key usable?
      Externals.GPG.Listkey(Key => ToStr(Key), Target => Key_Filename);
      Ada.Text_IO.Open(File => File_Handle,
                       Mode => Ada.Text_IO.In_File,
                       Name => Key_Filename);
      First_Line := Unbounded_Get_Line(File_Handle);
      Ada.Text_IO.Close(File_Handle);
      Process_Key(ToStr(First_Line), P, KP);
      if not Usable(KP) then
         -- It's not usable!
         Ada.Text_IO.Put_Line("Key " & ToStr(Key) & " is not usable!");
      elsif not Match_PR(KP, Role) then
         -- Doesn't match role.
         Ada.Text_IO.Put_Line("Key " & ToStr(Key) & " has wrong role!");
      else
         -- Is it a duplicate?
         Debug("Add_Key: Checking for duplicates");
         declare
            use type UBS;
         begin
            Debug("Add_Key: Approaching duplicates loop");
            for I in 1..List.Count loop
               Debug("Add_Key: In duplicates loop");
               Match := Match or Key = UAP.Value(List.KA, I);
            end loop;
         end;
         if not Match then
            -- Add the key.
            Debug("Add_Key: Adding a key");
            List.Count := List.Count + 1;
            UAP.Set(List.KA, List.Count, Key);
            Debug("Add_Key: Done adding a key");
         end if;
      end if;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Add_Key");
         raise;
   end Add_Key;

   -- Remove a key.  We then shuffle the array down.
   -- It will only remove the last key if there are multiple instances.
   -- The nastiness here is that we might be working with short
   --  fingerprints.  So choose the right hand end of the fingerprint.
   procedure Remove_Key (Key                  : in     UBS;
                         List                 : in out Key_List;
                         Exception_If_Missing : in     Boolean := False) is
      Key_Num    : Integer;
      Key_Found  : Boolean          := False;
      KS         : constant String  := ToStr(Key);
      Key_Length : constant Natural := KS'Length;
   begin
      for I in 1..List.Count loop
         declare
            -- Get this string...
            TK       : constant String := ToStr(UAP.Value(List.KA, I));
            TK_Start : Integer;
         begin
            -- Is TK longer?
            TK_Start := TK'First;
            if TK'Length > Key_Length then
               -- Advance TK_Start so that we've got matching lengths.
               TK_Start := TK'Last - Key_Length + 1;
            end if;
            if KS = TK(TK_Start..TK'Last) then
               Key_Found := True;
               Key_Num := I;
            end if;
         end;
      end loop;
      if Key_Found then
         -- Shuffle the array over it.
         for I in Key_Num .. List.Count - 1 loop
            UAP.Set(List.KA, I, UAP.Value(List.KA, I+1));
         end loop;
         -- Reduce the number of keys.
         List.Count := List.Count - 1;
      else
         if Exception_If_Missing then
            raise Key_Not_Found;
         end if;
      end if;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Remove_Key");
         raise;
   end Remove_Key;

   -- Turn the list of keys to send into `-r <key>' string.
   function Processed_Recipient_List (List : in Key_List) return String is
      Result : UBS := Ada.Strings.Unbounded.Null_Unbounded_String;
      use type UBS;
   begin
      for I in 1..List.Count loop
         Result := Result & ToUBS(" -r ") & UAP.Value(List.KA, I) & ToUBS(" ");
      end loop;
      return ToStr(Result);
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Processed_Recipient_List");
         raise;
   end Processed_Recipient_List;

   -- Get key fingerprint(s) for a key.  Add them.
   procedure Add_Keys_By_Fingerprint (Key_In : in     UBS;
                                      List   : in out Key_List;
                                      Role   : in     Key_Roles;
                                      Found  :    out Boolean) is
      K          : constant String := ToStr(Key_In);
      Tempfile   : constant String := Temp_File_Name("findkey-akbf");
      TFH        : Ada.Text_IO.File_Type;
      L          : UBS;
      Match      : Boolean := False;
   begin
      Debug("Add_Keys_By_Fingerprint (A): starting...");
      -- First, try finding some keys for this Key_In.
      Externals.GPG.Findkey(Key    => K,
                  Target => Tempfile);
      Debug("Add_Keys_By_Fingerprint (A): Externals.GPG.Findkey returned okay");
      -- Now, open the file, and read in each line as a key for Add_Key.
      Ada.Text_IO.Open(File => TFH,
                       Mode => Ada.Text_IO.In_File,
                       Name => Tempfile);
      Debug("Add_Keys_By_Fingerprint (A): opened file successfully");
  Fingerprint_Loop:
      loop
         begin
            L := Unbounded_Get_Line(TFH);
            Debug("Add_Keys_By_Fingerprint: Adding " & ToStr(L));
            Add_Key(L, List, Role);
            Match := True;
         exception
            when Ada.IO_Exceptions.End_Error =>
               exit Fingerprint_Loop;
         end;
      end loop Fingerprint_Loop;
      Ada.Text_IO.Close(TFH);
      Found := Match;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Add_Keys_By_Fingerprint (A)");
         raise;
   end Add_Keys_By_Fingerprint;

   procedure Add_Keys_By_Fingerprint (Key_In : in     UBS;
                                      List   : in out Key_List;
                                      Role   : in     Key_Roles) is
      Found : Boolean;
   begin
      Add_Keys_By_Fingerprint(Key_In, List, Role, Found);
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Add_Keys_By_Fingerprint (B)");
         raise;
   end Add_Keys_By_Fingerprint;

   -- Add the secret keys to this key list.
   procedure Add_Secret_Keys (Key_In : in     UBS;
                              List   : in out Key_List;
                              Role   : in     Key_Roles) is
      K          : constant String := ToStr(Key_In);
      Tempfile   : constant String := Temp_File_Name("findkey-ask");
      TFH        : Ada.Text_IO.File_Type;
      L          : UBS;
   begin
      -- First, try finding some _secret_ keys for this Key_In.
      Externals.GPG.Findkey_Secret(Key    => K,
                         Target => Tempfile);
      -- Now, open the file, and read in each line as a key for
      -- Add_Keys_By_Fingerprint (so we don't need to go hunting for the
      -- long codes.
      Ada.Text_IO.Open(File => TFH,
                       Mode => Ada.Text_IO.In_File,
                       Name => Tempfile);
  Fingerprint_Loop:
      loop
         begin
            L := Unbounded_Get_Line(TFH);
            Debug("Add_Keys_By_Fingerprint: Adding " & ToStr(L));
            Add_Keys_By_Fingerprint(L, List, Role);
         exception
            when Ada.IO_Exceptions.End_Error =>
               exit Fingerprint_Loop;
         end;
      end loop Fingerprint_Loop;
      Ada.Text_IO.Close(TFH);
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Add_Secret_Keys");
         raise;
   end Add_Secret_Keys;

   procedure Generate_Key_List (List      : in out Key_List;
                                Key_Heads : in out Key_Head_Array;
                                Key_Menu  : in out Menu_Array_Ptr) is
      C : Natural := 0;
   begin
      Debug("+Generate_Key_List");
      -- First, we generate a current list of info files for the keys.
      for I in 1..List.Count loop
         declare
            Key_File_Name : constant String
              := Temp_File_Name("key"
                                & Trim_Leading_Spaces(Integer'Image(I)));
         begin
            Externals.GPG.Listkey(Key    => ToStr(UAP.Value(List.KA, I)),
                        Target => Key_File_Name);
            -- Now, dump into our `head' array the first line of each.
            declare
               File_Handle    : Ada.Text_IO.File_Type;
               First_Line, P  : UBS;
               KP             : Key_Properties;
            begin
               Ada.Text_IO.Open(File => File_Handle,
                                Mode => Ada.Text_IO.In_File,
                                Name => Key_File_Name);
               First_Line := Unbounded_Get_Line(File_Handle);
               Ada.Text_IO.Close(File_Handle);
               Process_Key(ToStr(First_Line), P, KP);
--               if Usable(KP) then
                  C := C + 1;
                  declare
                     FL  : constant String := ToStr(P);
                     use Ada.Strings.Fixed;
                  begin
                     if FL'Length >= Key_Head_Length then
                        KHA.Set(Key_Heads, I, FL(FL'First..
                                                 FL'First+Key_Head_Length-1));
                     else
                        KHA.Set(Key_Heads, I, FL
                                & (Key_Head_Length - FL'Length) * ' ');
                     end if;
                  end;
--               end if;
            end;
         end;
      end loop;
      -- Now, create a menu with those key_heads.
      Key_Menu := new Keylist_Menus.MNA(1..C);
      for I in 1..C loop
         Key_Menu(I) := ToUBS(KHA.Value(Key_Heads, I));
      end loop;
      Debug("-Generate_Key_List");
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Generate_Key_List");
         raise;
   end Generate_Key_List;

   -- List the keys and edit them as appropriate (include removing a key
   -- from that list, and adding one from the keyring.
   procedure List_Keys (List : in out Key_List) is
      Key_Heads    : Key_Head_Array;
      Key_Menu     : Menu_Array_Ptr;
      Selection    : Keylist_Menus.CLM_Return;
      Selected     : Integer;
      SK_Selection : Specific_Key_Index;
   begin
      Debug("+List_Keys");
      KHA.Create(Key_Heads, Key_List_Initial_Size);
      Generate_Key_List(List, Key_Heads, Key_Menu);
  List_Keys_Loop:
      loop
         Ada.Text_IO.New_Line(5);
         Ada.Text_IO.Put_Line("** Key list menu:");
         Ada.Text_IO.Put_Line(Integer'Image(Keys.Count(List))
                         & " key(s) in key list      ");
         Selection := Keylist_Menu(Key_Menu.all);
         if not Selection.Is_Num then
            case Selection.I is
               when Done =>
                  exit List_Keys_Loop;
               when AddPattern => -- Add key from main keyring.
                  declare
                     New_Key : UBS;
                     use Ada.Text_IO;
                  begin
                     New_Line(1);
                     Put_Line("Main key ring access:");
                     New_Key
                       := ToUBS(Readline.Get_String("Type GPG search pattern: "));
                     Add_Keys_By_Fingerprint(New_Key, List, Encrypter);
                     Generate_Key_List(List, Key_Heads, Key_Menu);
                  end;
               when AddSearch => -- Do a search, then select a key.
                  declare
                     Search_List : Key_List;
                     Pattern     : UBS;
                     The_Key     : UBS;
                     Aborted     : Boolean;
                  begin
                     Empty_Keylist(Search_List);
                     Pattern := ToUBS(Readline.Get_String("Type GPG search pattern: "));
                     Add_Keys_By_Fingerprint(Pattern, Search_List, Encrypter);
                     Select_Key_From_List(Search_List, The_Key, Aborted);
                     if not Aborted then
                        Add_Key(The_Key, List, Encrypter);
                        Generate_Key_List(List, Key_Heads, Key_Menu);
                     end if;
                  end;
            end case;
         else
            -- Selection key menu.
            Selected := Selection.N;
        Specific_Key_Loop:
            loop
               Ada.Text_IO.New_Line(5);
               Ada.Text_IO.Put_Line("** Examining key currently in key list:");
               SK_Selection := Specific_Key_Menu1("Key: "
                                                  & ToStr(Key_Menu(Selected))
                                                  & NL
                                                  & "(d) Display details of key with less, (v) Verbosely"
                                                  & NL
                                                  & "(r) Remove key from list   (kql) Return to key list  "
                                                  & NL);
               case SK_Selection is
                  when Done =>
                     exit Specific_Key_Loop;
                  when Display =>
                     Externals.GPG.Viewkey(ToStr(UAP.Value(List.KA, Selected)));
                  when DisplayVerbose =>
                     Externals.GPG.Viewkey(ToStr(UAP.Value(List.KA, Selected)),
                                 Verbose => True);
                  when Remove =>
                     Remove_Key(UAP.Value(List.KA, Selected),
                                List);
                     Generate_Key_List(List, Key_Heads, Key_Menu);
                     exit Specific_Key_Loop;
                  when SSelect =>
                     Error("Menu should not have allowed SSelect here");
               end case;
            end loop Specific_Key_Loop;
         end if;
      end loop List_Keys_Loop;
      Debug("-List_Keys");
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.List_Keys");
         raise;
   end List_Keys;

   -- List the keys.  Either return with Aborted true, or
   -- The_Fingerprint set to the chosen fingerprint.
   procedure Select_Key_From_List (List            : in out Key_List;
                                   The_Fingerprint :    out UBS;
                                   Aborted         :    out Boolean) is
      Key_Heads        : Key_Head_Array;
      Key_Menu         : Menu_Array_Ptr;
      Selection        : Keylist_Menus.CLM_Return;
      Selected         : Integer;
      SK_Selection     : Specific_Key_Index;
      Abort_Real       : Boolean := False;
      Fingerprint_Real : UBS;
   begin
      Debug("+List_Keys");
      KHA.Create(Key_Heads, Key_List_Initial_Size);
      Generate_Key_List(List, Key_Heads, Key_Menu);
  List_Keys_Loop:
      loop
         Selection := Keylist_Menu2(Key_Menu.all);
         if not Selection.Is_Num then
            case Selection.I is
               when Done =>
                  Abort_Real := True;
                  exit List_Keys_Loop;
               when AddPattern =>
                  Error("Menu should not have allowed AddPattern here");
               when AddSearch =>
                  Error("Menu should not have allowed AddSearch here");
            end case;
         else
            -- Selection key menu.
            Selected := Selection.N;
        Specific_Key_Loop:
            loop
               Ada.Text_IO.New_Line(5);
               Ada.Text_IO.Put_Line("** Examining key found in search results:");
               SK_Selection := Specific_Key_Menu2("Key: "
                                                  & ToStr(Key_Menu(Selected))
                                                  & NL
                                                  & "(d) Display details of key with less, (v) Verbosely"
                                                  & NL
                                                  & "(s) Select this key         (kql) Return to key list  "
                                                  & NL);
               case SK_Selection is
                  when Done =>
                     exit Specific_Key_Loop;
                  when Display =>
                     Externals.GPG.Viewkey(ToStr(UAP.Value(List.KA, Selected)));
                  when DisplayVerbose =>
                     Externals.GPG.Viewkey(ToStr(UAP.Value(List.KA, Selected)),
                                 Verbose => True);
                  when SSelect =>
                    Fingerprint_Real := UAP.Value(List.KA, Selected);
                    exit List_Keys_Loop;
                  when Remove =>
                     Error("Menu should not have allowed Remove here");
                  end case;
               end loop Specific_Key_Loop;
            end if;
         end loop List_Keys_Loop;
      The_Fingerprint := Fingerprint_Real;
      Aborted := Abort_Real;
      Debug("-List_Keys");
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Select_Key_From_List");
         raise;
   end Select_Key_From_List;

   procedure First_Key_From_List (List            : in out Key_List;
                                  The_Fingerprint :    out UBS) is
   begin
      The_Fingerprint := UAP.Value(List.KA, 1);
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.First_Key_From_List");
         raise;
   end First_Key_From_List;



    -- Use keylist.
   procedure Use_Keylist (Recipients : in     UBS_Array;
                          List       : in out Key_List) is
      use Ada.Text_IO;
      Found_Key : array (Recipients'Range) of Boolean
        := (others => False);
   begin
      Debug("+Use_Keylist");
      -- New key list.
      -- Do something with the keylist.  Given a list of email addresses,
      -- open the relevant file (er, look in the array), and keep those key
      -- ids.
      for J in 1..Config.AKE_Count loop
         for I in Recipients'Range loop
            declare
               use type UBS;
               Found : Boolean;
            begin
               if UAP.Value(Config.AKE_Email, J) = Recipients(I) then
                  Add_Keys_By_Fingerprint(UAP.Value(Config.AKE_Key, J),
                                          List,
                                          Encrypter,
                                          Found);
                  Found_Key(I) := Found_Key(I) or Found;
               end if;
            end;
         end loop;
      end loop;
      -- Now, run through the recipients that we _haven't_ found a key for.
      for I in Recipients'Range loop
         if Found_Key(I) then
            Put_Line("Info: Added keylist key(s) for recipient: `"
                     & ToStr(Recipients(I)) & "'");
         else
            declare
               Search_List : Key_List;
               use type UBS;
            begin
               Empty_Keylist(Search_List);
               Add_Keys_By_Fingerprint(Recipients(I), Search_List, Encrypter);
               -- Remove any on the XK list.  Dead easy: simply run
               -- through the entire XK list saying `Remove Key'.
               for J in 1 .. Config.XK_Count loop
                  Remove_Key(UAP.Value(Config.XK_Key, J), Search_List);
               end loop;
               if Count(Search_List) > 0 then
                  Put_Line("Info: Adding non-keylist key(s) for recipient: `"
                           & ToStr(Recipients(I)) & "'");
                  -- Add the keys in Search_List to List.
                  for J in 1 .. Search_List.Count loop
                     Add_Key(UAP.Value(Search_List.KA, J), List, Encrypter);
                  end loop;
               else
                  Put_Line("Warning: Cannot find key for recipient: `"
                           & ToStr(Recipients(I)) & "'");
               end if;
            end;
         end if;
      end loop;
      Debug("-Use_Keylist");
   exception
      when Ada.IO_Exceptions.Name_Error =>
         Put_Line("Warning: Can't open keylist file");
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Use_Keylist");
         raise;
   end Use_Keylist;

   procedure Empty_Keylist (List : in out Key_List) is
   begin
      UAP.Create(List.KA, Key_List_Initial_Size);
      List.Count := 0;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Empty_Keylist");
         raise;
   end;

   function Count (List : in Key_List) return Natural is
   begin
      return List.Count;
   exception
      when others =>
         Ada.Text_IO.Put_Line(Ada.Text_IO.Standard_Error,
                              "Exception raised in Keys.Count");
         raise;
   end Count;

end Keys;


